@@ -90,29 +90,46 @@ def _remove(s, t):
     assert i >= 0
     return s[:i] + s[i+len(t):]
 
-def hash_headers(hasher, canonicalize_headers, headers, include_headers,
-                 sigheaders, sig):
-    """Sign message header fields."""
+def select_headers(headers, include_headers):
+    """Select message header fields to be signed/verified.
+
+    >>> h = [('from','biz'),('foo','bar'),('from','baz'),('subject','boring')]
+    >>> i = ['from','subject','to','from']
+    >>> select_headers(h,i)
+    [('from', 'baz'), ('subject', 'boring'), ('from', 'biz')]
+    >>> h = [('From','biz'),('Foo','bar'),('Subject','Boring')]
+    >>> i = ['from','subject','to','from']
+    >>> select_headers(h,i)
+    [('From', 'biz'), ('Subject', 'Boring')]
+    """
     sign_headers = []
     lastindex = {}
     for h in include_headers:
+        assert h == h.lower()
         i = lastindex.get(h, len(headers))
         while i > 0:
             i -= 1
-            if h.lower() == headers[i][0].lower():
+            if h == headers[i][0].lower():
                 sign_headers.append(headers[i])
                 break
         lastindex[h] = i
+    return sign_headers
+
+def hash_headers(hasher, canonicalize_headers, headers, include_headers,
+                 sigheaders, sig):
+    """Update hash for signed message header fields."""
+    sign_headers = select_headers(headers,include_headers)
     # The call to _remove() assumes that the signature b= only appears
     # once in the signature header
     cheaders = canonicalize_headers.canonicalize_headers(
         [(sigheaders[0][0], _remove(sigheaders[0][1], sig[b'b']))])
-    sign_headers += [(x[0], x[1].rstrip()) for x in cheaders]
-    for x in sign_headers:
-        hasher.update(x[0])
+    # the dkim sig is hashed with no trailing crlf, even if the
+    # canonicalization algorithm would add one.
+    for x,y in sign_headers + [(x, y.rstrip()) for x,y in cheaders]:
+        hasher.update(x)
         hasher.update(b":")
-        hasher.update(x[1])
-
+        hasher.update(y)
+    return sign_headers
 
 def validate_signature_fields(sig):
     """Validate DKIM-Signature fields.
@@ -164,10 +181,8 @@ def rfc822_parse(message):
     """Parse a message in RFC822 format.
 
     @param message: The message in RFC822 format. Either CRLF or LF is an accepted line separator.
-
-    @return Returns a tuple of (headers, body) where headers is a list of (name, value) pairs.
+    @return: Returns a tuple of (headers, body) where headers is a list of (name, value) pairs.
     The body is a CRLF-separated string.
-
     """
     headers = []
     lines = re.split(b"\r?\n", message)
@@ -193,7 +208,17 @@ def rfc822_parse(message):
 
 
 def fold(header):
-    """Fold a header line into multiple crlf-separated lines at column 72."""
+    """Fold a header line into multiple crlf-separated lines at column 72.
+
+    >>> fold(b'foo')
+    'foo'
+    >>> fold(b'foo  '+b'foo'*24).splitlines()[0]
+    'foo  '
+    >>> fold(b'foo'*25).splitlines()[-1]
+    ' foo'
+    >>> len(fold(b'foo'*25).splitlines()[0])
+    72
+    """
     i = header.rfind(b"\r\n ")
     if i == -1:
         pre = b""
@@ -204,19 +229,61 @@ def fold(header):
     while len(header) > 72:
         i = header[:72].rfind(b" ")
         if i == -1:
-            j = i
+            j = 72
         else:
             j = i + 1
-        pre += header[:i] + b"\r\n "
+        pre += header[:j] + b"\r\n "
         header = header[j:]
     return pre + header
 
+#: Hold messages and options during DKIM signing and verification.
 class DKIM(object):
+  # NOTE - the first 2 indentation levels are 2 instead of 4 
+  # to minimize changed lines from the function only version.
 
-  def __init__(self,message,logger=None,signature_algorithm=b'rsa-sha256'):
-    (self.headers, self.body) = rfc822_parse(message)
-    self.domain = None
-    self.selector = 'default'
+  #: The U{RFC5322<http://tools.ietf.org/html/rfc5322#section-3.6>}
+  #: complete list of singleton headers (which should
+  #: appear at most once).  This can be used for a "paranoid" or
+  #: "strict" signing mode.
+  #: Bcc in this list is in the SHOULD NOT sign list, the rest could
+  #: be in the default FROZEN list, but that could also make signatures 
+  #: more fragile than necessary.  
+  #: @since: 0.5
+  RFC5322_SINGLETON = ('date','from','sender','reply-to','to','cc','bcc',
+        'message-id','in-reply-to','references')
+
+  #: Header fields to protect from additions by default.
+  #: 
+  #: The short list below is the result more of instinct than logic.
+  #: @since: 0.5
+  FROZEN = ('from','date','subject')
+
+  #: The rfc4871 recommended header fields to sign
+  #: @since: 0.5
+  SHOULD = (
+    'sender', 'reply-to', 'subject', 'date', 'message-id', 'to', 'cc',
+    'mime-version', 'content-type', 'content-transfer-encoding', 'content-id',
+    'content- description', 'resent-date', 'resent-from', 'resent-sender',
+    'resent-to', 'resent-cc', 'resent-message-id', 'in-reply-to', 'references',
+    'list-id', 'list-help', 'list-unsubscribe', 'list-subscribe', 'list-post',
+    'list-owner', 'list-archive'
+  )
+
+  #: The rfc4871 recommended header fields not to sign.
+  #: @since: 0.5
+  SHOULD_NOT = (
+    'return-path', 'received', 'comments', 'keywords', 'bcc', 'resent-bcc',
+    'dkim-signature'
+  )
+
+  #: Create a DKIM instance to sign and verify rfc5322 messages.
+  #:
+  #: @param message: an RFC822 formatted message to be signed or verified
+  #: (with either \\n or \\r\\n line endings)
+  #: @param logger: a logger to which debug info will be written (default None)
+  #: @param signature_algorithm: the signing algorithm to use when signing
+  def __init__(self,message=None,logger=None,signature_algorithm=b'rsa-sha256'):
+    self.set_message(message)
     if logger is None:
         logger = get_default_logger()
     self.logger = logger
@@ -224,7 +291,98 @@ class DKIM(object):
         raise ParameterError(
             "Unsupported signature algorithm: "+signature_algorithm)
     self.signature_algorithm = signature_algorithm
+    #: Header fields which should be signed.  Default from RFC4871
+    self.should_sign = set(DKIM.SHOULD)
+    #: Header fields which should not be signed.  The default is from RFC4871.
+    #: Attempting to sign these headers results in an exception.
+    #: If it is necessary to sign one of these, it must be removed
+    #: from this list first.
+    self.should_not_sign = set(DKIM.SHOULD_NOT)
+    #: Header fields to sign an extra time to prevent additions.
+    self.frozen_sign = set(DKIM.FROZEN)
 
+  def add_frozen(self,s):
+    """ Add headers not in should_not_sign to frozen_sign.
+    @param s: list of headers to add to frozen_sign
+    @since: 0.5
+
+    >>> dkim = DKIM()
+    >>> dkim.add_frozen(DKIM.RFC5322_SINGLETON)
+    >>> sorted(dkim.frozen_sign)
+    ['cc', 'date', 'from', 'in-reply-to', 'message-id', 'references', 'reply-to', 'sender', 'subject', 'to']
+    """
+    self.frozen_sign.update(x.lower() for x in s
+        if x.lower() not in self.should_not_sign)
+
+  #: Load a new message to be signed or verified.
+  #: @param message: an RFC822 formatted message to be signed or verified
+  #: (with either \\n or \\r\\n line endings)
+  #: @since: 0.5
+  def set_message(self,message):
+    if message:
+      self.headers, self.body = rfc822_parse(message)
+    else:
+      self.headers, self.body = [],''
+    #: The DKIM signing domain last signed or verified
+    self.domain = None
+    #: The DKIM key selector last signed or verified
+    self.selector = 'default'
+    #: The list of headers last signed or verified.  Each header
+    #: is a name,value tuple.  FIXME: The headers are canonicalized.
+    #: This could be more useful as original headers.
+    self.signed_headers = []
+
+  def default_sign_headers(self):
+    """Return the default list of headers to sign: those in should_sign or
+    frozen_sign, with those in frozen_sign signed an extra time to prevent
+    additions.
+    @since: 0.5"""
+    hset = self.should_sign | self.frozen_sign
+    include_headers = [ x for x,y in self.headers
+        if x.lower() in hset ]
+    return include_headers + [ x for x in include_headers
+        if x.lower() in self.frozen_sign]
+
+  def all_sign_headers(self):
+    """Return header list of all existing headers not in should_not_sign.
+    @since: 0.5"""
+    return [x for x,y in self.headers if x.lower() not in self.should_not_sign]
+
+  #: Sign an RFC822 message and return the DKIM-Signature header line.
+  #:
+  #: The include_headers option gives full control over which header fields
+  #: are signed.  Note that signing a header field that doesn't exist prevents
+  #: that field from being added without breaking the signature.  Repeated
+  #: fields (such as Received) can be signed multiple times.  Instances
+  #: of the field are signed from bottom to top.  Signing a header field more
+  #: times than are currently present prevents additional instances
+  #: from being added without breaking the signature.
+  #:
+  #: The length option allows the message body to be appended to by MTAs
+  #: enroute (e.g. mailing lists that append unsubscribe information)
+  #: without breaking the signature.
+  #:
+  #: The default include_headers for this method differs from the backward
+  #: compatible sign function, which signs all headers not 
+  #: in should_not_sign.  The default list for this method can be modified 
+  #: by tweaking should_sign and frozen_sign (or even should_not_sign).
+  #: It is only necessary to pass an include_headers list when precise control
+  #: is needed.
+  #:
+  #: @param selector: the DKIM selector value for the signature
+  #: @param domain: the DKIM domain value for the signature
+  #: @param privkey: a PKCS#1 private key in base64-encoded text form
+  #: @param identity: the DKIM identity value for the signature
+  #: (default "@"+domain)
+  #: @param canonicalize: the canonicalization algorithms to use
+  #: (default (Simple, Simple))
+  #: @param include_headers: a list of strings indicating which headers
+  #: are to be signed (default rfc4871 recommended headers)
+  #: @param length: true if the l= tag should be included to indicate
+  #: body length signed (default False).
+  #: @return: DKIM-Signature header field terminated by '\r\n'
+  #: @raise DKIMException: when the message, include_headers, or key are badly
+  #: formed.
   def sign(self, selector, domain, privkey, identity=None,
         canonicalize=(b'simple',b'simple'), include_headers=None, length=False):
     try:
@@ -240,10 +398,17 @@ class DKIM(object):
     headers = canon_policy.canonicalize_headers(self.headers)
 
     if include_headers is None:
-        include_headers = [x[0].lower() for x in headers]
-    else:
-        include_headers = [x.lower() for x in include_headers]
-    sign_headers = [x for x in headers if x[0].lower() in include_headers]
+        include_headers = self.default_sign_headers()
+
+    # rfc4871 says FROM is required
+    if 'from' not in ( x.lower() for x in include_headers ):
+        raise ParameterError("The From header field MUST be signed")
+
+    # raise exception for any SHOULD_NOT headers, call can modify 
+    # SHOULD_NOT if really needed.
+    for x in include_headers:
+        if x.lower() in self.should_not_sign:
+            raise ParameterError("The %s header field SHOULD NOT be signed"%x)
 
     body = canon_policy.canonicalize_body(self.body)
 
@@ -262,26 +427,21 @@ class DKIM(object):
         (b'q', b"dns/txt"),
         (b's', selector),
         (b't', str(int(time.time())).encode('ascii')),
-        (b'h', b" : ".join(x[0] for x in sign_headers)),
+        (b'h', b" : ".join(include_headers)),
         (b'bh', bodyhash),
         (b'b', b""),
     ] if x]
+    include_headers = [x.lower() for x in include_headers]
+    # record what verify should extract
+    self.include_headers = tuple(include_headers)
 
     sig_value = fold(b"; ".join(b"=".join(x) for x in sigfields))
-    dkim_header = canon_policy.canonicalize_headers([
-        [b'DKIM-Signature', b' ' + sig_value]])[0]
-    # the dkim sig is hashed with no trailing crlf, even if the
-    # canonicalization algorithm would add one.
-    if dkim_header[1][-2:] == b'\r\n':
-        dkim_header = (dkim_header[0], dkim_header[1][:-2])
-    sign_headers.append(dkim_header)
-
-    self.logger.debug("sign headers: %r" % sign_headers)
+    dkim_header = (b'DKIM-Signature', b' ' + sig_value)
     h = hashlib.sha256()
-    for x in sign_headers:
-        h.update(x[0])
-        h.update(b":")
-        h.update(x[1])
+    sig = dict(sigfields)
+    self.signed_headers = hash_headers(
+        h, canon_policy, headers, include_headers, [dkim_header],sig)
+    self.logger.debug("sign headers: %r" % self.signed_headers)
 
     try:
         sig2 = RSASSA_PKCS1_v1_5_sign(
@@ -294,20 +454,27 @@ class DKIM(object):
     self.selector = selector
     return b'DKIM-Signature: ' + sig_value + b"\r\n"
 
+  #: Verify a DKIM signature.
+  #: @type idx: int
+  #: @param idx: which signature to verify.  The first (topmost) signature is 0.
+  #: @type dnsfunc: callable
+  #: @param dnsfunc: an option function to lookup TXT resource records
+  #: for a DNS domain.  The default uses dnspython or pydns.
+  #: @return: True if signature verifies or False otherwise
+  #: @raise DKIMException: when the message, signature, or key are badly formed
+  def verify(self,idx=0,dnsfunc=get_txt):
 
-  def verify(self,dnsfunc=get_txt):
-
-    sigheaders = [x for x in self.headers if x[0].lower() == b"dkim-signature"]
-    if len(sigheaders) < 1:
+    sigheaders = [(x,y) for x,y in self.headers if x.lower() == b"dkim-signature"]
+    if len(sigheaders) <= idx:
         return False
 
-    # Currently, we only validate the first DKIM-Signature line found.
+    # By default, we validate the first DKIM-Signature line found.
     try:
-        sig = parse_tag_value(sigheaders[0][1])
+        sig = parse_tag_value(sigheaders[idx][1])
     except InvalidTagValueList as e:
         raise MessageFormatError(e)
 
-    sig = parse_tag_value(sigheaders[0][1])
+    sig = parse_tag_value(sigheaders[idx][1])
     logger = self.logger
     logger.debug("sig: %r" % sig)
 
@@ -353,12 +520,21 @@ class DKIM(object):
         raise KeyFormatError(e)
     try:
         pk = parse_public_key(base64.b64decode(pub[b'p']))
+    except KeyError:
+        raise KeyFormatError("incomplete public key: %s" % s)
     except (TypeError,UnparsableKeyError) as e:
         raise KeyFormatError("could not parse public key (%s): %s" % (pub[b'p'],e))
-
-    include_headers = re.split(br"\s*:\s*", sig[b'h'])
+    include_headers = [x.lower() for x in re.split(br"\s*:\s*", sig[b'h'])]
+    self.include_headers = tuple(include_headers)
+    # address bug#644046 by including any additional From header
+    # fields when verifying.  Since there should be only one From header,
+    # this shouldn't break any legitimate messages.  This could be
+    # generalized to check for extras of other singleton headers.
+    if 'from' in include_headers:
+      include_headers.append('from')      
     h = hasher()
-    hash_headers(h, canon_policy, headers, include_headers, sigheaders, sig)
+    self.signed_headers = hash_headers(
+        h, canon_policy, headers, include_headers, sigheaders, sig)
     try:
         signature = base64.b64decode(re.sub(br"\s+", b"", sig[b'b']))
         return RSASSA_PKCS1_v1_5_verify(
@@ -371,27 +547,29 @@ def sign(message, selector, domain, privkey, identity=None,
          signature_algorithm=b'rsa-sha256',
          include_headers=None, length=False, logger=None):
     """Sign an RFC822 message and return the DKIM-Signature header line.
-
     @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
     @param selector: the DKIM selector value for the signature
     @param domain: the DKIM domain value for the signature
     @param privkey: a PKCS#1 private key in base64-encoded text form
     @param identity: the DKIM identity value for the signature (default "@"+domain)
     @param canonicalize: the canonicalization algorithms to use (default (Simple, Simple))
-    @param include_headers: a list of strings indicating which headers are to be signed (default all headers)
+    @param include_headers: a list of strings indicating which headers are to be signed (default all headers not listed as SHOULD NOT sign)
     @param length: true if the l= tag should be included to indicate body length (default False)
     @param logger: a logger to which debug info will be written (default None)
+    @return: DKIM-Signature header field terminated by \\r\\n
+    @raise DKIMException: when the message, include_headers, or key are badly formed.
     """
 
     d = DKIM(message,logger=logger)
+    if not include_headers:
+        include_headers = d.all_sign_headers()
     return d.sign(selector, domain, privkey, identity=identity, canonicalize=canonicalize, include_headers=include_headers, length=length)
 
 def verify(message, logger=None, dnsfunc=get_txt):
-    """Verify a DKIM signature on an RFC822 formatted message.
-
+    """Verify the first (topmost) DKIM signature on an RFC822 formatted message.
     @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
     @param logger: a logger to which debug info will be written (default None)
-
+    @return: True if signature verifies or False otherwise
     """
     d = DKIM(message,logger=logger)
     try:
